from pathlib import Path
from PIL import Image, ImageDraw

from skitso.atom import Container, Point


class Scene(Container):

    def __init__(self, canvas_size, base_folder_path, color="black", antialias=False):
        super().__init__(position=Point(0, 0))
        self.color = color
        self.width, self.height = canvas_size
        self.create_canvas()
        self.build_folder(base_folder_path)
        self.next_tick_id = 1
        self.antialias = antialias

    def build_folder(self, base_folder_path):
        self.folder_path = (
            Path(base_folder_path) / type(self).__name__.lower() / "frames"
        )
        self.folder_path.mkdir(parents=True, exist_ok=True)

    def create_canvas(self):
        self.image = Image.new("RGB", (self.width, self.height), self.color)
        self.draw = ImageDraw.Draw(self.image)
        self.draw.image = self.image  # type: ignore
        self.draw.fontmode = "L"

    def tick(self):
        # need to create the image from scratch
        self.create_canvas()
        for item in self.children:
            item.draw_me(self.draw)

        new_img_path = self.folder_path / f"{self.next_tick_id:08}.jpg"
        if self.antialias:
            # only known way to get antialiasing: resize up and down.
            im = self.image.resize(
                (self.width * 2, self.height * 2), resample=Image.Resampling.LANCZOS
            )
            im = im.resize((self.width, self.height), resample=Image.Resampling.LANCZOS)
        else:
            im = self.image
        im.save(new_img_path, subsampling=0, quality=95)
        self.next_tick_id += 1



from dataclasses import dataclass
from enum import Enum


@dataclass
class Point:
    x: float
    y: float

    def __add__(self, other):
        if isinstance(other, Point):
            ox = other.x
            oy = other.y
        elif isinstance(other, Vector):
            ox = other.dx
            oy = other.dy
        elif isinstance(other, tuple) and len(other) == 2:
            ox, oy = other
        else:
            raise ValueError(f"Cannot add {type(other)} to Point")
        return Point(self.x + ox, self.y + oy)

    def __sub__(self, other):
        if not isinstance(other, Point):
            raise ValueError(f"Cannot substract {type(other)} to Point")
        return Vector(self.x - other.x, self.y - other.y)

    def __iter__(self):
        # allows easy conversion to tuple
        yield self.x
        yield self.y


@dataclass
class Vector:
    dx: float
    dy: float

    def __add__(self, other):
        if isinstance(other, Vector):
            return Vector(self.dx + other.dx, self.dy + other.dy)
        else:
            raise ValueError(f"Cannot add {type(other)} to Vector")

    def __mul__(self, num):
        if not isinstance(num, (int, float)):
            raise ValueError(f"Cannot multiply Vector by {type(num)}")
        return Vector(self.dx * num, self.dy * num)

    def __iter__(self):
        # allows easy conversion to tuple
        yield self.dx
        yield self.dy


LEFT = Vector(-1, 0)
RIGHT = Vector(1, 0)
UP = Vector(0, -1)
DOWN = Vector(0, 1)


class AlignmentDial(Enum):
    LOW = -10
    HALF = 0
    HIGH = 10


class Aligment:
    def __init__(self, horizontal, vertical):
        self.horizontal = horizontal
        self.vertical = vertical


TOP_EDGE = Aligment(None, AlignmentDial.LOW)
BOTTOM_EDGE = Aligment(None, AlignmentDial.HIGH)
CENTER_CENTER = Aligment(AlignmentDial.HALF, AlignmentDial.HALF)
LEFT_EDGE = Aligment(AlignmentDial.LOW, None)
RIGHT_EDGE = Aligment(AlignmentDial.HIGH, None)


class Movable:
    # Every Movable must have a position that represents:
    # - it's current location (starts at 0, 0 which is the top-left corner of the screen)
    #
    # Also, it has an "end" property that returns the lowest-right corner of the Movable,
    # such property MUST be dynamically computed based on the Movable's position, so if position
    # changes, the "end" value changes as well.
    #
    # Together, position and end makes possible the "align_to" method.

    def __init__(self, position, *args, **kwargs) -> None:
        if not isinstance(position, Point):
            raise ValueError(f"Expected Point, got {type(position)}")
        self.position = position

    @property
    def end(self):
        raise NotImplementedError('Not implemented "end" for %s' % self.__class__.__name__)

    @property
    def box_height(self):
        return self.end.y - self.position.y

    @property
    def box_width(self):
        return self.end.x - self.position.x

    def move_to(self, point):
        if not isinstance(point, Point):
            raise ValueError(f"Expected Point, got {type(point)}")
        self.position = point

    def shift(self, vector):
        if not isinstance(vector, Vector):
            raise ValueError(f"Expected Vector, got {type(vector)}")
        self.position += vector

    def move_end_to(self, point):
        if not isinstance(point, Point):
            raise ValueError(f"Expected Point, got {type(point)}")
        # Given that "end" it's a property, we need to move the position in such a way that
        # the end of the Movable is at the requested point
        current_end = self.end
        delta = point - current_end
        self.position += delta
        assert self.end == point

    def to_edge(self, other, alignment):
        # Makes coincide the edge of "self" with requested edges of "other"
        if not isinstance(other, Movable):
            raise ValueError(f"Expected Movable, got {type(other)}")
        if not isinstance(alignment, Aligment):
            raise ValueError(f"Expected Aligment, got {type(alignment)}")

        current_end = self.end
        if alignment.vertical is None:
            new_y = self.position.y  # leave y unchanged
        elif alignment.vertical == AlignmentDial.LOW:  # top edge
            new_y = other.position.y
        elif alignment.vertical == AlignmentDial.HIGH:  # bottom edge
            new_y = other.end.y - current_end.y
        else:  # vertical center
            other_delta_y = other.end.y - other.position.y
            self_delta_y = current_end.y - self.position.y
            new_y = other.position.y + (other_delta_y - self_delta_y) / 2

        if alignment.horizontal is None:
            new_x = self.position.x  # leave x unchanged
        elif alignment.horizontal == AlignmentDial.LOW:  # left edge
            new_x = other.position.x
        elif alignment.horizontal == AlignmentDial.HIGH:  # right edge
            new_x = other.end.x - current_end.x
        else:  # horizontal center
            other_delta_x = other.end.x - other.position.x
            self_delta_x = current_end.x - self.position.x
            new_x = other.position.x + (other_delta_x - self_delta_x) / 2

        self.move_to(Point(new_x, new_y))

    def center_respect_to(self, other):
        self.to_edge(other, CENTER_CENTER)



import math
from PIL import Image, ImageFont, ImageDraw

from skitso.atom import BaseImgElem, Container, Point


class Rectangle(BaseImgElem):
    def __init__(
        self,
        x,
        y,
        width,
        height,
        fill_color,
        fill_opacity=None,
        stroke_color=None,
        stroke_width=None,
    ):
        super().__init__(Point(x, y))
        self.width = width
        self.height = height
        self.fill_color = fill_color
        self.fill_opacity = fill_opacity
        self.stroke_color = stroke_color
        self.stroke_width = stroke_width

    @property  # needed for movable API
    def end(self):
        return Point(self.x + self.width, self.y + self.height)

    @property
    def x(self):
        return self.position.x

    @property
    def y(self):
        return self.position.y

    def x1y1x2y2(self):
        pfrom, pto = self.position, self.end
        return pfrom.x, pfrom.y, pto.x, pto.y

    def _draw_kw(self):
        kw = {}
        kw["fill"] = self.fill_color
        if self.stroke_width is not None and self.stroke_color is not None:
            kw["width"] = self.stroke_width
            kw["outline"] = self.stroke_color
        return kw

    def draw_me(self, pencil):
        kw = self._draw_kw()
        pencil.rectangle(self.x1y1x2y2(), **kw)


class RoundedRectangle(Rectangle):
    def __init__(
        self,
        x,
        y,
        width,
        height,
        fill_color,
        fill_opacity=None,
        stroke_color=None,
        stroke_width=None,
        corner_radius=None,
    ):
        super().__init__(
            x, y, width, height, fill_color, fill_opacity, stroke_color, stroke_width
        )
        self.corner_radius = corner_radius

    def _draw_kw(self):
        kw = super()._draw_kw()
        kw["radius"] = self.corner_radius
        return kw

    def draw_me(self, pencil):
        kw = self._draw_kw()
        pencil.rounded_rectangle(self.x1y1x2y2(), **kw)


class Line(BaseImgElem):
    def __init__(self, x1, y1, x2, y2, color, thickness):
        # Lets compute the inserion point: the top-leftmost point of (not the line)
        # but of the wrapping rectangle around the line
        position = Point(min(x1, x2), min(y1, y2))
        super().__init__(position)
        self.relative_end = Point(max(x1, x2), max(y1, y2)) - position
        self.line_from = Point(x1 - position.x, y1 - position.y)
        self.line_to = Point(x2 - position.x, y2 - position.y)
        self.color = color
        self.thickness = thickness

    @property
    def end(self):
        return self.position + self.relative_end

    def draw_me(self, pencil):
        shift = self.position
        x1, y1 = shift + self.line_from
        x2, y2 = shift + self.line_to
        pencil.line((x1, y1, x2, y2), fill=self.color, width=self.thickness)

    def get_slope(self, in_degrees=False):
        # 0 is the horizontal axis to the right. Pi (or 180°) its the opposite
        # Pi/2 it's the vertical going up, and 3/2 * Pi going down.
        delta = self.line_from - self.line_to
        slope = math.atan2(delta.dy, delta.dx)
        if in_degrees:
            slope = slope * 180 / math.pi
        return slope

    def get_length(self):
        delta = self.line_from - self.line_to
        return (delta.dx**2 + delta.dy**2) ** (1 / 2)


class ArrowTip(BaseImgElem):
    def __init__(self, x, y, height, angle, color):
        # Arrow tip should be poiting exactly at x,y and the direction
        # should be the same as the vector
        # Arrow tip height is measured over such direction
        self.color = color
        self.height = height
        self.base = height  # can be changed later
        self.tip_angle = angle

        vertices = self.compute_vertices(x, y)
        min_x = min(v.x for v in vertices)
        min_y = min(v.y for v in vertices)
        super().__init__(Point(min_x, min_y))
        self.relative_vertices = [Point(v.x - min_x, v.y - min_y) for v in vertices]
        self.relative_end = Point(
            max(v.x for v in vertices) - min_x, max(v.y for v in vertices) - min_y
        )

    @property
    def end(self):
        return self.position + self.relative_end

    def draw_me(self, pencil):
        shift = self.position
        positioned_triangle = [vtx + shift for vtx in self.relative_vertices]
        raw = [(vtx.x, vtx.y) for vtx in positioned_triangle]
        pencil.polygon(raw, fill=self.color)

    def compute_vertices(self, tip_x, tip_y):
        half_base = self.base / 2
        alpha = math.atan(half_base / self.height) * 180 / math.pi
        hypothenuse = (self.height**2 + half_base**2) ** (1 / 2)
        v0 = Point(tip_x, tip_y)
        vertices = [v0]
        for beta in [self.tip_angle - alpha, self.tip_angle + alpha]:
            beta_radians = beta * math.pi / 180
            vertex_x = math.cos(beta_radians) * hypothenuse
            vertex_y = math.sin(beta_radians) * hypothenuse
            vertices.append(v0 + Point(vertex_x, vertex_y))
        return vertices


class Arrow(BaseImgElem):
    def __init__(self, x1, y1, x2, y2, color, thickness, tip_height):
        self.line = Line(x1, y1, x2, y2, color, thickness)
        # ArrowTip is at (x2, y2)
        # Arrow tip height is measured over the direction of the line
        angle = self.line.get_slope(in_degrees=True)
        self.tip = ArrowTip(x2, y2, tip_height, angle, color)
        self.tip_delta = self.tip.position - self.position

        # Line is drawn from (x1, y1) to (x2, y2). But if thickness != 1
        # line will be shortened a bit to not provoke a weird broken pointy arrow
        if thickness > 1:
            # we will truncate half tip_height.
            chopped_length = tip_height / 2
            angle_radians = angle * math.pi / 180
            chopped_x = math.cos(angle_radians) * chopped_length
            chopped_y = math.sin(angle_radians) * chopped_length
            self.truncated_line = Line(
                x1, y1, x2 + chopped_x, y2 + chopped_y, color, thickness
            )
        else:
            self.truncated_line = None

    @property
    def position(self):
        return self.line.position

    @position.setter
    def position(self, value):
        self.line.position = value
        self.tip.position = value + self.tip_delta
        if self.truncated_line is not None:
            self.truncated_line.position = value

    @property
    def end(self):
        return self.line.end

    @property
    def color(self):
        return self.line.color

    @property
    def thickness(self):
        return self.line.thickness

    def draw_me(self, pencil):
        self.tip.draw_me(pencil)
        if self.truncated_line is not None:
            self.truncated_line.draw_me(pencil)
        else:
            self.line.draw_me(pencil)


class DeadArrow(Container):
    def __init__(self, x, y, length, tip_height, color, thickness):
        self.color = color
        self.thickness = thickness
        position = Point(x, y)
        elbow = Point(x + length, y)
        end = Point(x + length, y + length)
        self.line = Line(x, y, elbow.x, elbow.y, color, thickness)
        self.arrow = Arrow(elbow.x, elbow.y, end.x, end.y, color, thickness, tip_height)
        super().__init__(position)
        self.add(self.line)
        self.add(self.arrow)


class Text(BaseImgElem):
    def __init__(
        self, x, y, text, font_name, font_size, color="white", align="left", bold=False
    ):
        super().__init__(Point(x, y))
        self.text = text
        self.font_name = font_name
        self.font_size = font_size
        self.bold = bold
        self.color = color
        self.align = align
        self.font = self.load_font()

    def load_font(self):
        return ImageFont.load_default(self.font_size)

    @property
    def end(self):
        if not hasattr(self, "relative_end"):
            temp_img = Image.new("RGBA", (1000, 1000))
            temp_draw = ImageDraw.Draw(temp_img)
            box = temp_draw.textbbox((0, 0), self.text, font=self.font)
            self.relative_end = Point(box[2], box[3])
        return self.position + self.relative_end

    def draw_me(self, pencil):
        font = self.load_font()
        x, y = self.position
        pencil.text((x, y), self.text, font=font, fill=self.color, align=self.align)
